#include<cstdio>
#include<cmath>
#include<opencv/cv.h>
#include<opencv/cxcore.h>
#include<opencv/highgui.h>
#include"graph.h"
#include"graph.cpp"
#include"maxflow.cpp"

#define SMOOTH_COST 1
#define COLOR_BAND 256
#define DISPARITY_MAX 16
#define SAD_SIZE 5
#define CORRECTION 16

typedef Graph<float,float,float> GraphType;

int left_right(int col, int row1, int row2, CvSize size) {
	return size.width*size.height+col*(size.width-1)+row1;
}

int top_down(int col1, int col2, int row, CvSize size) {
	return size.width*size.height+(size.width-1)*size.height+col1*size.width+row;
}

IplImage* calculate_sad(IplImage *image) {
	unsigned char *temp;
	float *tempSAD;
	float SAD_SQUARE=(SAD_SIZE*2+1)*(SAD_SIZE*2+1);
	float num;
	int width,height;
	int i,j;

	CvSize size=cvGetSize(image);
	IplImage *SAD=cvCreateImage(size, IPL_DEPTH_32F, 1);
	cvSetZero(SAD);

	for(height=SAD_SIZE; height<size.height-SAD_SIZE; height++) {
		for(width=SAD_SIZE; width<size.width-SAD_SIZE; width++) {
			num=0;
			for(j=-SAD_SIZE; j<=SAD_SIZE; j++) {
				temp = &((unsigned char *)(image->imageData + image->widthStep*(height+j)))[width];

				for(i=-SAD_SIZE; i<=SAD_SIZE; i++) {
					num+=temp[i];
				}
			}

			tempSAD = &((float *)(SAD->imageData + SAD->widthStep*height))[width];
			tempSAD[0] = num/SAD_SQUARE;
		}
	}

	return(SAD);
}

IplImage* hierarchical(IplImage *leftImage, IplImage *rightImage) {
	GraphType *g;
 	int	currentNode;
	unsigned char *temp_fp;
	float	*tempLeftSAD, *tempRightSAD;
	int x,y;
	int j,k;
	int DENOMI, K_MAX, MAX_LABEL;

	for(j=1, DENOMI=2; j<DISPARITY_MAX; j++) {
		DENOMI*=2;
		if(DENOMI > DISPARITY_MAX) break;
	}

	MAX_LABEL=j;
	int **LABEL = (int **)malloc(sizeof(int*)*MAX_LABEL);
	LABEL[0] = (int*)malloc(sizeof(int));
	LABEL[0][0] = 0;
	LABEL[1] = (int*)malloc(sizeof(int));
	LABEL[1][0] = DISPARITY_MAX/2;

	for(j=1, DENOMI=2; j<MAX_LABEL; j++) {
		DENOMI*=2;
		K_MAX = DENOMI/4;

		LABEL[j*2] = (int *)malloc(sizeof(int)*K_MAX);
		for(k=0; k<K_MAX; k++) {
			LABEL[j*2][k] = (1+4*k)*DISPARITY_MAX/DENOMI;
		}

		LABEL[j*2+1] = (int *)malloc(sizeof(int)*K_MAX);
		for(k=0; k<K_MAX; k++) {
			LABEL[j*2+1][k] = (3+4*k)*DISPARITY_MAX/DENOMI;
		}
	}

	for(j=0; j<MAX_LABEL*2; j++) {
		if(j<2) K_MAX = 1;
		else if(j%2==0) K_MAX = (int)(pow(2.0,j/2-1));
		else K_MAX = (int)(pow(2.0,(j-1)/2-1));
	}

	CvSize size = cvGetSize(leftImage);

	IplImage *imageLeftSAD;
	IplImage *imageRightSAD;
	imageLeftSAD=calculate_sad(leftImage);
	imageRightSAD=calculate_sad(rightImage);

	IplImage *fpImage = cvCreateImage(size, IPL_DEPTH_8U, 1);
	cvSetZero(fpImage);

	double LARGE_NUM=10000000.0;
	int success=0;
	int alpha,alpha_p, alpha_q;
	float weightSource,weightSink;
	float weightLeft2Alpha,weightAlpha2Left;
	float weightRight2Alpha,weightAlpha2Right;
	float weightTop2Alpha,weightAlpha2Top;
	float weightDown2Alpha,weightAlpha2Down;
	int lrNodeNum,tdNodeNum;

	do {
		success=0;
		for(alpha=0; alpha<MAX_LABEL*2; alpha++) {
			if(alpha<2) K_MAX = 1;
			else if(alpha%2==0) K_MAX = (int)(pow(2.0,alpha/2-1));
			else K_MAX = (int)(pow(2.0,(alpha-1)/2-1));

			g = new GraphType(size.width*size.height+(size.width-1)*size.height+size.width*(size.height-1),(size.width-1)*size.height*2+size.width*(size.height-1)*2); 
			g -> add_node(size.width*size.height+(size.width-1)*size.height+size.width*(size.height-1)); 
			for(y=SAD_SIZE, currentNode=0; y<size.height-SAD_SIZE; y++) {
				tempLeftSAD  = &((float*)(imageLeftSAD->imageData + imageLeftSAD->widthStep*y))[DISPARITY_MAX];
				tempRightSAD = &((float*)(imageRightSAD->imageData + imageRightSAD->widthStep*y))[DISPARITY_MAX];
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];

				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE; x++) {
					weightSource = abs(tempLeftSAD[x]-tempRightSAD[x-temp_fp[x]]);
					for(k=0,alpha_p=100; k<K_MAX; k++) {
						if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) {
							alpha_p=LABEL[alpha][k];
						}
					}

					weightSink = abs(tempLeftSAD[x]-tempRightSAD[x-alpha_p]);
					g -> add_tweights(currentNode,weightSource,weightSink);
					currentNode++;
				}
			}

			for(y=SAD_SIZE; y<size.height-SAD_SIZE; y++) {
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];

				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE-1; x++) {
					if(temp_fp[x] == temp_fp[x+1]) {
						for(k=0,alpha_p=100; k<K_MAX; k++)
							if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) alpha_p=LABEL[alpha][k];
						weightLeft2Alpha = (float)(SMOOTH_COST * abs(temp_fp[x] - alpha_p));
						lrNodeNum = y*(size.width-DISPARITY_MAX*2)+x;
						g -> add_edge( lrNodeNum, lrNodeNum+1, weightLeft2Alpha, weightLeft2Alpha);
					}
					else {
						for(k=0,alpha_p=100; k<K_MAX; k++)
							if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) alpha_p=LABEL[alpha][k];

						for(k=0,alpha_q=100; k<K_MAX; k++)
							if(alpha_q > abs(temp_fp[x+1]-LABEL[alpha][k])) alpha_q=LABEL[alpha][k];

 						weightSource = (float)(SMOOTH_COST * abs(temp_fp[x] - temp_fp[x+1]));
						weightSink   = (float)(SMOOTH_COST * abs(alpha_p - alpha_q));
						g -> add_tweights( currentNode,  weightSource, weightSink);

						if(weightSource <= weightSink) {
							weightLeft2Alpha  = (float)(SMOOTH_COST * (abs(alpha_p   - temp_fp[x]) - weightSource));
							weightRight2Alpha = (float)(SMOOTH_COST * (abs(temp_fp[x] - alpha_q) - weightSource));

							weightAlpha2Left =  10000;
							weightAlpha2Right = 10000;

							if(0 > weightLeft2Alpha) weightLeft2Alpha = 0;
							if(0 > weightRight2Alpha) weightRight2Alpha = 0;
						}
						else {
							weightAlpha2Left  = (float)(SMOOTH_COST * (abs(temp_fp[x]   - alpha_q) - weightSink));
							weightAlpha2Right = (float)(SMOOTH_COST * (abs(alpha_p - temp_fp[x+1]) - weightSink));

							weightLeft2Alpha =  10000;
							weightRight2Alpha = 10000;

							if(0 > weightAlpha2Left) weightAlpha2Left = 0;
							if(0 > weightAlpha2Right) weightAlpha2Right = 0;
						}

						lrNodeNum = y*(size.width-DISPARITY_MAX*2)+x;

						g -> add_edge(  lrNodeNum, currentNode, weightLeft2Alpha,  weightAlpha2Left );
						g -> add_edge( currentNode, lrNodeNum+1, weightAlpha2Right, weightRight2Alpha );

						currentNode++;
					}
				}
			}

			for(y=SAD_SIZE; y<size.height-SAD_SIZE-1; y++) {
				temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[DISPARITY_MAX];

				for(x=0; x<size.width-DISPARITY_MAX-SAD_SIZE; x++) {
					if(temp_fp[x] == temp_fp[x+size.width]) {
						for(k=0,alpha_p=100; k<K_MAX; k++) {
							if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) {
								alpha_p=LABEL[alpha][k];
							}
						}
						weightTop2Alpha = (float)(SMOOTH_COST * abs(temp_fp[x]   - alpha_p));
						tdNodeNum = y*(size.width-DISPARITY_MAX*2)+x;

						g -> add_edge( tdNodeNum, tdNodeNum+size.width, weightTop2Alpha, weightTop2Alpha);

					} else {
						for(k=0,alpha_p=100; k<K_MAX; k++)
							if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) alpha_p=LABEL[alpha][k];

						for(k=0,alpha_q=100; k<K_MAX; k++) 
							if(alpha_q > abs(temp_fp[x+size.width]-LABEL[alpha][k])) alpha_q=LABEL[alpha][k];

						weightSource = (float)(SMOOTH_COST * abs(temp_fp[x] - temp_fp[x+size.width]));
						weightSink   = (float)(SMOOTH_COST * abs(alpha_p - alpha_q));

						g -> add_tweights( currentNode, weightSource, weightSink);
						if(weightSource <= weightSink) {
							weightTop2Alpha = (float)(SMOOTH_COST * (abs(alpha_p   - temp_fp[x+size.width]) - weightSource));
							weightDown2Alpha = (float)(SMOOTH_COST * (abs(temp_fp[x] - alpha_q) - weightSource));
							weightAlpha2Top =  10000;
							weightAlpha2Down = 10000;

							if(0 > weightTop2Alpha) weightTop2Alpha = 0;
							if(0 > weightDown2Alpha) weightDown2Alpha = 0;
						}
						else {
							weightAlpha2Top = (float)(SMOOTH_COST * (abs(temp_fp[x]   - alpha_q) - weightSink));
							weightAlpha2Down = (float)(SMOOTH_COST * (abs(alpha_p - temp_fp[x+size.width]) - weightSink));

							weightTop2Alpha =  10000;
							weightDown2Alpha = 10000;
							if(0 > weightAlpha2Top) weightAlpha2Top = 0;
							if(0 > weightAlpha2Down) weightAlpha2Down = 0;
						}

						tdNodeNum = y*(size.width-DISPARITY_MAX*2)+x;

						g -> add_edge( tdNodeNum, currentNode, weightTop2Alpha, weightAlpha2Top );
						g -> add_edge( currentNode, tdNodeNum+size.width, weightAlpha2Down, weightDown2Alpha );

						currentNode++;
					}
				}
			}

			double large_num = g -> maxflow();

			if(large_num < LARGE_NUM) {
				success = 1;
				LARGE_NUM = large_num;

				for(y=SAD_SIZE, currentNode=0; y<size.height-SAD_SIZE; y++) {
					temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[0];

					for(x=DISPARITY_MAX; x<size.width-SAD_SIZE; x++) {
						if (g->what_segment(currentNode) == GraphType::SOURCE) {
							for(k=0,alpha_p=100; k<K_MAX; k++)
								if(alpha_p > abs(temp_fp[x]-LABEL[alpha][k])) alpha_p=LABEL[alpha][k];
							temp_fp[x] = alpha_p;
						}

						currentNode++;
					}
				}
			}

			delete g; 
		}
	}
	while(success);

	printf("Flow = %lf\n", LARGE_NUM);

	for(y=SAD_SIZE; y<size.height-SAD_SIZE; y++) {
		temp_fp = &((uchar*)(fpImage->imageData + fpImage->widthStep*y))[0];

		for(x=DISPARITY_MAX; x<size.width-SAD_SIZE; x++) {
			if(temp_fp[x]*CORRECTION > 255) temp_fp[x] = 255;
			else temp_fp[x] *= CORRECTION;
		}
	}

	return(fpImage);
}

int main(int argc, char *argv[]){
	char left[]  = "t_img01.jpg";
	char right[] = "t_img03.jpg";
	char dst[] = "dst.jpg";
	char window_left[] = "LEFT";
	char window_right[]= "RIGHT";
	char window_parallax[]= "PARALLAX";
	cvNamedWindow(window_left,CV_WINDOW_AUTOSIZE);
	cvNamedWindow(window_right,CV_WINDOW_AUTOSIZE);
	cvNamedWindow(window_parallax,CV_WINDOW_AUTOSIZE);

	IplImage *left_image = cvLoadImage(left,0);
	IplImage *right_image  = cvLoadImage(right,0);
	IplImage *parallax_image = hierarchical(left_image,right_image);

	cvShowImage(window_left,left_image);
	cvShowImage(window_right,right_image);
	cvShowImage(window_parallax,parallax_image);

	cvSaveImage(dst,parallax_image);

	cvWaitKey(0);
	cvDestroyAllWindows();

	return 0;
}
